//
//  C++ implementation of the F90 interface of the Data Subpsace
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
# include <stdio.h>
# include <DataSubspace.h>
# include <f90cxx.h>
# include <Handle.h>
# include <stringutil.h>
# include <cstring>

F90MODPROCDEF(Handle<DataSubspace>,dssinter,newdatasubspace)() {
  return new DataSubspace();
}


F90MODPROCDEF(Handle<DataSubspace>,dssinter,getdatasubspace) (
	     const Handle<Block> &inBlockH
	     ) {
  return new DataSubspace(inBlockH.addressOfObject());
}


F90MODPROCDEF(void,dssinter,dssread) (
	      Handle<DataSubspace> &outDssH,
	      const Handle<Block>  &inBlockH
	      ) {
  inBlockH.addressOfObject() >> *outDssH.addressOfObject();
}


F90MODPROCDEF(void,dssinter,dsswrite) (
	      Handle<DataSubspace> &inDssH,
	      const Handle<Block>  &outBlockH
	      ) {
  outBlockH.addressOfObject() << *inDssH.addressOfObject();
}


F90MODPROCDEF(void,dssinter,dssrelease) (
	      Handle<DataSubspace> &dssH
	      ) {
  delete dssH.addressOfObject();
}


F90MODPROCDEF(Handle<DScomp>,dssinter,dssaddcomp) (
              const Handle<DataSubspace> &dssH
	      ) {
  return dssH.addressOfObject()->addComponent();
}


F90MODPROCDEF(void,dssinter,dssdeletecomponent) (
              const Handle<DataSubspace> &dssH,
              int &iComp
	      ) {
  dssH.addressOfObject()->deleteComponent(iComp);
}


F90MODPROCDEF(Handle<DScomp>,dssinter,dsscomponent) (
	      const Handle<DataSubspace> &dssH,
              int &iComp
	      ) {
  return dssH.addressOfObject()->component(iComp);
}


F90MODPROCDEF(int,dssinter,dsscomponents) (
	      const Handle<DataSubspace> &dssH
	      ) {
  return dssH.addressOfObject()->components();
}


F90MODPROCDEF(void,dssinter,subspaceselectexpression) (
              char *value,
              int valueLength,
	      const Handle<DataSubspace> &dssH,
              char *axes, // optional
              int axesLength	      
	      ) {
  // Axes gets passed as a comma delimited string of axis names

  DSS::DSaxes dsaxes;

  if( axes ){
    std::string parseString = trim(std::string(axes,axesLength));

    while( parseString.size() != 0 ){
      std::string::size_type icomma = parseString.find(",");
      if( icomma == std::string::npos ) icomma = parseString.size();
      dsaxes.push_back( parseString.substr(0,icomma) );
      parseString.erase(0,icomma+1);
    }
  }

  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dssH.addressOfObject()->selectExpression(dsaxes).c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(dssH.addressOfObject()->selectExpression(dsaxes).size() ), valueLength );
  memcpy( value,dssH.addressOfObject()->selectExpression(dsaxes).c_str(), len );
  memset( value + len, ' ', valueLength - len );
}


F90MODPROCDEF(void,dssinter,compselectexpression) (
              char *value,
              int valueLength,
	      const Handle<DScomp> &dscompH,
              char *axes, // optional
              int axesLength	     
	      ) {
  // Axes gets passed as a comma delimited string of axis names

  DSS::DSaxes dsaxes;

  if( axes ){
    std::string parseString = trim(std::string(axes,axesLength));

    while( parseString.size() != 0 ){
      std::string::size_type icomma = parseString.find(",");
      if( icomma == std::string::npos ) icomma = parseString.size();
      dsaxes.push_back( parseString.substr(0,icomma) );
      parseString.erase(0,icomma+1);
    }
  }
  
  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dscompH.addressOfObject()->selectExpression(dsaxes).c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(dscompH.addressOfObject()->selectExpression(dsaxes).size() ), valueLength );
  memcpy( value,dscompH.addressOfObject()->selectExpression(dsaxes).c_str(), len );
  memset( value + len, ' ', valueLength - len );
}


F90MODPROCDEF(void,dssinter,filterselectexpression) (
              char *value,
              int valueLength,
	      const Handle<DSfilter> &dsfiltH
	      ) {
  //sprintf(value,
  //  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dsfiltH.addressOfObject()->selectExpression().c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(dsfiltH.addressOfObject()->selectExpression().size() ), valueLength );
  memcpy( value,dsfiltH.addressOfObject()->selectExpression().c_str(), len );
  memset( value + len, ' ', valueLength - len );
}


F90MODPROCDEF(void,dssinter,getsubspaceranges) (
	      const Handle<DataSubspace> &dssH,
	      const char *axis,
	      Dope<int,1>*   lowerB, 
	      Dope<double,1>* lower,
	      Dope<double,1>* upper,
	      Dope<int,1>*   upperB,
	      int axisLength
	      ) {
  std::string axisName = trim(std::string(axis,axisLength));
  std::list<Range<double> > lRanges = dssH.addressOfObject()->ranges(axisName);

  uint32_t size = lRanges.size();

  double *lLower = new double[size];
  double *lUpper = new double[size];
  int    *lLowerB = new int[size];
  int    *lUpperB = new int[size];

  RangeBound tempLowerB, tempUpperB;

  std::list<Range<double> >::iterator irange;

  unsigned i;
  for(irange=lRanges.begin(), i=0; irange!=lRanges.end(); irange++, i++){
    irange->values(lLower[i], lUpper[i], tempLowerB, tempUpperB);
    lLowerB[i] = tempLowerB;
    lUpperB[i] = tempUpperB;
  }

  Dope<int,1> llb(lLowerB, &size);
  Dope<double,1> ll(lLower, &size);
  Dope<double,1> lu(lUpper, &size);
  Dope<int,1> lub(lUpperB, &size);

  *lowerB = llb;
  *lower  = ll;
  *upper  = lu;
  *upperB = lub;
}

  
F90MODPROCDEF(void,dssinter,getcomponentranges) (
	      const Handle<DScomp> &dscompH,
	      const char *axis,
	      Dope<int,1>*   lowerB, 
	      Dope<double,1>* lower,
	      Dope<double,1>* upper,
	      Dope<int,1>*   upperB,
	      int axisLength
	      ) {
  std::string axisName = trim(std::string(axis,axisLength));
  std::list<Range<double> > lRanges = dscompH.addressOfObject()->ranges(axisName);

  uint32_t size = lRanges.size();

  double *lLower = new double[size];
  double *lUpper = new double[size];
  int    *lLowerB = new int[size];
  int    *lUpperB = new int[size];

  RangeBound tempLowerB, tempUpperB;

  std::list<Range<double> >::iterator irange;

  unsigned i;
  for(irange=lRanges.begin(), i=0; irange!=lRanges.end(); irange++, i++){
    irange->values(lLower[i], lUpper[i], tempLowerB, tempUpperB);
    lLowerB[i] = tempLowerB;
    lUpperB[i] = tempUpperB;
  }

  Dope<int,1> llb(lLowerB, &size);
  Dope<double,1> ll(lLower, &size);
  Dope<double,1> lu(lUpper, &size);
  Dope<int,1> lub(lUpperB, &size);

  *lowerB = llb;
  *lower  = ll;
  *upper  = lu;
  *upperB = lub;
}

  
F90MODPROCDEF(void,dssinter,getfilterranges) (
	      const Handle<DSfilter> &dsfilterH,
	      Dope<int,1>*   lowerB, 
	      Dope<double,1>* lower,
	      Dope<double,1>* upper,
	      Dope<int,1>*   upperB
	      ) {
  std::list<Range<double> > lRanges = dsfilterH.addressOfObject()->ranges();

  uint32_t size = lRanges.size();

  double *lLower = new double[size];
  double *lUpper = new double[size];
  int    *lLowerB = new int[size];
  int    *lUpperB = new int[size];

  RangeBound tempLowerB, tempUpperB;

  std::list<Range<double> >::iterator irange;

  unsigned i;
  for(irange=lRanges.begin(), i=0; irange!=lRanges.end(); irange++, i++){
    irange->values(lLower[i], lUpper[i], tempLowerB, tempUpperB);
    lLowerB[i] = tempLowerB;
    lUpperB[i] = tempUpperB;
  }

  Dope<int,1> llb(lLowerB, &size);
  Dope<double,1> ll(lLower, &size);
  Dope<double,1> lu(lUpper, &size);
  Dope<int,1> lub(lUpperB, &size);

  *lowerB = llb;
  *lower  = ll;
  *upper  = lu;
  *upperB = lub;
}

  
F90MODPROCDEF(void,dssinter,releasedssranges) (
	      Dope<int,1>*   lowerB, 
	      Dope<double,1>* lower,
	      Dope<double,1>* upper,
	      Dope<int,1>*   upperB
	      ) {
  delete lowerB->addr;
  delete lower->addr;
  delete upper->addr;
  delete upperB->addr;
}


F90MODPROCDEF(void,dssinter,dssdeletefilter) (
              Handle<DScomp>  &dscompH,
	      int             &filterNumber
	      ) {
  dscompH.addressOfObject()->deleteFilter((unsigned)filterNumber);
}


F90MODPROCDEF(Handle<DSfilter>,dssinter,dssfilterbynumber) (
              Handle<DScomp>  &dscompH,
	      int             &filterNumber
	      ) {
  return dscompH.addressOfObject()->filter((unsigned)filterNumber);
}


F90MODPROCDEF(Handle<DSfilter>,dssinter,dssfilterbyname) (
	      Handle<DScomp>             &dscompH,
 	      char *                     name,
 	      DSS::FilterType            &filterType,
 	      int                        nameLength
 	      ) {
  std::string filterName = trim(std::string(name,nameLength));
  return dscompH.addressOfObject()->filter(filterName, filterType);
}


F90MODPROCDEF(bool,dssinter,dsshasfilter) (
 	      Handle<DScomp>             &dscompH,
 	      char *                     name,
 	      DSS::FilterType            &filterType,
 	      int                        nameLength
 	      ) {
  std::string filterName = trim(std::string(name,nameLength));
  return dscompH.addressOfObject()->hasFilter(filterName, filterType);
}


F90MODPROCDEF(int,dssinter,dssfilters) (
	      Handle<DScomp>  &dscompH
	      ) {
  return dscompH.addressOfObject()->filters();
}


F90MODPROCDEF(int,dssinter,dssfiltertype) (
	      Handle<DSfilter> &dsfilterH
	      ) {
  return dsfilterH.addressOfObject()->type();
}


F90MODPROCDEF(void,dssinter,dssfiltername) (
	      char *           value,
	      int              valueLength,
	      Handle<DSfilter> &dsfilterH
	      ) {
  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dsfilterH.addressOfObject()->name().c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int( dsfilterH.addressOfObject()->name().size() ), valueLength );
  memcpy( value, dsfilterH.addressOfObject()->name().c_str(), len );
  memset( value + len, ' ', valueLength - len );

}


F90MODPROCDEF(void,dssinter,dssfilterunits) (
	      char *            value,
	      int               valueLength,
	      Handle<DSfilter> &dsfilterH
	      ) {
  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dsfilterH.addressOfObject()->units().c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(  dsfilterH.addressOfObject()->units().size() ), valueLength );
  memcpy( value,  dsfilterH.addressOfObject()->units().c_str(), len );
  memset( value + len, ' ', valueLength - len );

}


F90MODPROCDEF(void,dssinter,dssfiltercomps) (
	      char *            value,
	      int               valueLength,
	      Handle<DSfilter> &dsfilterH
	      ) {
  DSS::DSfilterComp filtComps = dsfilterH.addressOfObject()->comps();
  std::string sComps = filtComps[0] + "," + filtComps[1];

  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  sComps.c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(  sComps.size() ), valueLength );
  memcpy( value, sComps.c_str(), len );
  memset( value + len, ' ', valueLength - len );

}


F90MODPROCDEF(void,dssinter,dssfilterblock) (
	      char *            value,
	      int               valueLength,
	      Handle<DSfilter> &dsfilterH
	      ) {

  //sprintf(value,
  //	  "%-*.*s",
  //	  valueLength,
  //	  valueLength,
  //	  dsfilterH.addressOfObject()->blockSpec().c_str() );

  // AI: Notice that F90 is asking for 256 char len and the index  
  // in C++ starts in 0, while in F90 starts in 1... So therefore, we have to pass
  // to F90 255 char len... Or we have to do what it's written below..
  const int len = min( int(  dsfilterH.addressOfObject()->blockSpec().size() ), valueLength );
  memcpy( value, dsfilterH.addressOfObject()->blockSpec().c_str(), len );
  memset( value + len, ' ', valueLength - len );
}


F90MODPROCDEF(void,dssinter,dssfiltermask) (
	      Handle<DSfilter> &dsfilterH,
	      Dope<uint32_t,1> * onBits,
	      Dope<uint32_t,1> * offBits
	      ) {
     vector<uint32_t> onvec = dsfilterH.addressOfObject()->onBitMasks();
     vector<uint32_t> offvec = dsfilterH.addressOfObject()->offBitMasks();

     uint32_t size = onvec.size();
     uint32_t *onarray = new uint32_t[size];
     uint32_t *offarray = new uint32_t[size];

     for(unsigned i=0; i<size; i++){
       onarray[i] = onvec[i];
       offarray[i] = offvec[i];
     }

     Dope<uint32_t,1> ondope(onarray, &size);
     Dope<uint32_t,1> offdope(offarray, &size);

     *onBits = ondope;
     *offBits = offdope;
}

F90MODPROCDEF(void,dssinter,dssmerge) (
	      Handle<DataSubspace> &dss1H,
	      Handle<DataSubspace> &dss2H
              ) {
  *(dss1H.addressOfObject()) *= *(dss2H.addressOfObject());
  return;
}

F90MODPROCDEF(void,dssinter,dssdelete) (
	      Handle<Block> &blockH
              ) {
   deleteDSS( blockH.addressOfObject() );
  return;
}
